/* David Leonard, 2002. Public domain. */

/* Debugging tools to show information about a USB device */

#include <stdio.h>
#include <err.h>
#include <sys/ioctl.h>
#include <dev/usb/usb.h>

#define hexval(c) ((c) >= 'a' ? (c)-'a'+10 : (c)-'0')
/* return substrings of "\x00Fred,\x01Bob,\x02Mary" */
static char *
desct(desc, v)
	char *desc;
	unsigned char v;
{
	static char buf[1024];
	char *b = buf, *p = desc;

	while ((hexval(p[0])<<4)|hexval(p[1]) != v) {
		p += 2;
		if (*p != ' ')
			errx(1, "desct '%s' %d", desc, p - desc);
		p++;
		while (*p && *p != ' ')
			p++;
		if (*p == '\0')
			return "?";
		p++;
	}
	p += 2;
	if (*p != ' ')
		errx(1, "desct '%s' %d", desc, p - desc);
	p++;
	while (*p && *p != ' ')
		*b++ = *p++;
	*b = '\0';
	return buf;
}

/* Return bit choices from "\x01foo,\x02bar,\x04baz" */
static char *
descm(desc, v)
	char *desc;
	unsigned char v;
{
	static char buf[1024];
	char *b = buf, *p = desc;

	while (*p) {
		unsigned char m = (hexval(p[0])<<4)|hexval(p[1]);
		unsigned char match = (v & m);
		p += 2;
		if (*p != ' ')
			errx(1, "descm '%s' %d", desc, p - desc);
		p++;
		if (match) {
			if (b == buf) *b++ = '<';
			else *b++ = ',';
		}
		while (*p && *p != ' ') {
			if (match)
				*b++ = *p;
			p++;
		}
		if (*p == ' ')
			p++;
	}
	if (b == buf)
		*b++ = '<';
	*b++ = '>';
	*b++ = '\0';
	return buf;
}

/* Dump information about a ugen device */

void
dumpusb(ugenfd)
	int ugenfd;
{
#define PL(i,t,n)   printf("%*s%s %d\n", i*4, "", t, n)
#define PB(i,p,f)   printf("%*s%-20s  0x%02x\n", i*4, "", #f, (p).f)
#define PD(i,p,f)   printf("%*s%-20s%6u\n", i*4, "", #f, (p).f)
#define PW(i,p,f)   printf("%*s%-20s0x%04x\n", i*4, "", #f, UGETW((p).f))
#define PM(i,p,f,d) printf("%*s%-20s  0x%02x %s\n", i*4, "", #f, (p).f, \
				descm(d, (p).f))
#define PT(i,p,f,d) printf("%*s%-20s  0x%02x %s\n", i*4, "", #f, (p).f, \
				desct(d, (p).f))
#define PX(i,p,f)   printf("%*s%-20s  0x%02x ", i*4, "", #f, (p).f)

#define DESCTYPE "01 device 02 config 04 interface 05 endpoint"

	int config, interface, alt, endpoint;
	usb_device_descriptor_t devdesc;

	printf("ugen device on fd %d:\n", ugenfd);

	if (ioctl(ugenfd, USB_GET_DEVICE_DESC, &devdesc) == -1)
		err(1, "USB_GET_DEVICE_DESC");

/*
	PB(0, devdesc, bLength);
	PT(0, devdesc, bDescriptorType, DESCTYPE);
*/
	PW(0, devdesc, bcdUSB);
	PB(0, devdesc, bDeviceClass);
	PB(0, devdesc, bDeviceSubClass);
	PB(0, devdesc, bDeviceProtocol);
	PB(0, devdesc, bMaxPacketSize);
	PW(0, devdesc, idVendor);
	PW(0, devdesc, idProduct);
	PW(0, devdesc, bcdDevice);
	PD(0, devdesc, iManufacturer);
	PD(0, devdesc, iProduct);
	PD(0, devdesc, iSerialNumber);
	PD(0, devdesc, bNumConfigurations);
	printf("\n");

	for (config = 0; config < devdesc.bNumConfigurations; config++)
	{
	    struct usb_config_desc cd;

	    PL(0, "config", config);
	    cd.config_index = config;
	    if (ioctl(ugenfd, USB_GET_CONFIG_DESC, &cd) == -1)
		err(1, "USB_GET_CONFIG_DESC");
	/*
	    PB(1, cd.desc, bLength);
	    PT(1, cd.desc, bDescriptorType, DESCTYPE);
	*/
	    PW(1, cd.desc, wTotalLength);
	    PD(1, cd.desc, bNumInterface);
	    PB(1, cd.desc, bConfigurationValue);
	    PB(1, cd.desc, iConfiguration);
	    PM(1, cd.desc, bmAttributes,
		"20 remote-wakeup 40 self-powered 80 bus-powered");
	    PD(1, cd.desc, bMaxPower);
	    printf("\n");

	    for (interface = 0; interface < cd.desc.bNumInterface; interface++)
	    {
		struct usb_interface_desc ifc;
		struct usb_alt_interface alti;
		int alt, endpt;

		PL(1, "interface", interface);
		alti.config_index = config;
		alti.interface_index = interface;
	        if (ioctl(ugenfd, USB_GET_ALTINTERFACE, &alti) == -1)
		    err(1, "USB_GET_ALTINTERFACE");

		PD(2, alti, alt_no);
		alt = alti.alt_no;

		ifc.config_index = config;
		ifc.interface_index = interface;
		ifc.alt_index = alt;
	        if (ioctl(ugenfd, USB_GET_INTERFACE_DESC, &ifc) == -1)
		    err(1, "USB_GET_INTERFACE_DESC");

	/* 
		PB(2, ifc.desc, bLength);
		PT(2, ifc.desc, bDescriptorType, DESCTYPE);
	*/
		PB(2, ifc.desc, bInterfaceNumber);
		PB(2, ifc.desc, bAlternateSetting);
		PB(2, ifc.desc, bNumEndpoints);
		PB(2, ifc.desc, bInterfaceClass);
		PB(2, ifc.desc, bInterfaceSubClass);
		PB(2, ifc.desc, bInterfaceProtocol);
		PD(2, ifc.desc, iInterface);
		printf("\n");

		for (endpoint=0;endpoint<ifc.desc.bNumEndpoints;endpoint++)
		{
		    struct usb_endpoint_desc ed;

		    PL(2, "endpoint", endpoint);

		    ed.config_index = config;
		    ed.interface_index = interface;
		    ed.alt_index = alt;
		    ed.endpoint_index = endpoint;
	            if (ioctl(ugenfd, USB_GET_ENDPOINT_DESC, &ed) == -1)
		        err(1, "USB_GET_ENDPOINT_DESC");

	/*
		    PB(3, ed.desc, bLength);
		    PT(3, ed.desc, bDescriptorType, DESCTYPE);
	*/
		    PX(3, ed.desc, bEndpointAddress);
		    printf("%s addr %x\n", 
			UE_GET_DIR(ed.desc.bEndpointAddress) == UE_DIR_IN ? 
					"in" : "out",
			UE_GET_ADDR(ed.desc.bEndpointAddress));
		    PT(3, ed.desc, bmAttributes,
			"00 control 02 bulk 03 interrupt"
			" 05 iso-async 09 iso-adapt 0d iso-sync");
		    PW(3, ed.desc, wMaxPacketSize);
		    PB(3, ed.desc, bInterval);
		}
	    }
	}
}
