/* David Leonard, 2002. Public domain. */

/*
 * Search for an iPAQ attached via USB, initialise it, and then
 * provide a pseudo-tty interface for a program such as pppd.
 */

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <err.h>
#include <util.h>
#include <sys/ioctl.h>
#include <sys/poll.h>
#include <sys/types.h>
#include <sys/param.h>
#include <dev/usb/usb.h>

#define USB_VENDOR_COMPAQ 0x049f
#define USB_PRODUCT_IPAQ  0x0003

int debug = 0;

extern void dumpusb(int fd);

/* Continuously copy data between two pairs of file descriptors */
void
relay(ra, wa, rb, wb)
{
	struct pollfd pfd[2], *p;
	int aready, bready;
	int i, len, wlen;
	char buf[65536];

	if (debug)
		fprintf(stderr, "ra %d wa %d rb %d wb %d\n", ra, wa, rb, wb);

	for (;;) {
		pfd[0].fd = ra;
		pfd[0].events = POLLIN;
		pfd[0].revents = 0;
		pfd[1].fd = rb;
		pfd[1].events = POLLIN;
		pfd[1].revents = 0;

		if (poll(pfd, 2, -1) == -1) {
			warn("poll");
			break;
		}
		if ((pfd[0].revents & POLLIN) != 0) {
			if ((len = read(ra, buf, sizeof buf)) == -1) {
			    warn("read ra");
			    break;
			}

			if (debug) {
			    fprintf(stderr, "read %d from ra %d:", len, ra);
			    for (i = 0; i < len; i++)
				    if (isprint(buf[i]) && buf[i] != '\\')
					    fputc(buf[i], stderr);
				    else
					    fprintf(stderr, "\\%03o", buf[i]);
			    fputc('\n', stderr);
			}

			if (len == 0)
			    break;
			if ((wlen = write(wb, buf, len)) == -1) {
			    warn("write wb");
			    break;
			}
			if (debug)
			    fprintf(stderr, "wrote %d to wb %d\n", wlen, wb);
			if (wlen != len)
			    fprintf(stderr, "short write\n");
		}
		if ((pfd[1].revents & POLLIN) != 0) {
			if ((len = read(rb, buf, sizeof buf)) == -1) {
				warn("read rb");
				break;
			}

			if (debug) {
			    fprintf(stderr, "read %d from rb %d:", len, rb);
			    for (i = 0; i < len; i++)
				    if (isprint(buf[i]) && buf[i] != '\\')
					    fputc(buf[i], stderr);
				    else
					    fprintf(stderr, "\\%03o", buf[i]);
			    fputc('\n', stderr);
			}

			if (len == 0)
			    break;
			if ((wlen = write(wa, buf, len)) == -1) {
			    warn("write wa");
			    break;
			}
			if (debug)
			    fprintf(stderr, "wrote %d to wa %d\n", wlen, wa);
			if (wlen != len)
			    fprintf(stderr, "short write\n");
		}
	}
}

int
main(argc, argv)
	int argc;
	char *argv[];
{
	char path[MAXPATHLEN];
	char buf[1024];
	char *ipaqdev;
	int usbfd, ugenfd, infd, outfd;
	int ch, i, flags, error;
	int pty, errfd;
	pid_t child;
	struct usb_device_info udi;
	struct usb_ctl_request req;
	extern int optind;
	extern char *optarg;

	/* Process command line arguments */
	error = 0;
	udi.addr = 0;
	while ((ch = getopt(argc, argv, "a:d")) != -1) 
		switch (ch) {
		case 'a':
			udi.addr = atoi(optarg);
			break;
		case 'd':
			debug = 1;
			break;
		default:
			error = 1;
		}
	if (argc == optind)
		error = 1;
	if (error) {
		fprintf(stderr, "usage: %s"
		    " [-a addr]"
		    " [-d]"
		    " cmd args...\n", argv[0]);
		exit(1);
	}

	/* Access the USB subsystem */
	usbfd = -1;
	for (i = 0; usbfd == -1; i++) {
		snprintf(path, sizeof path, "/dev/usb%d", i);
		usbfd = open(path, O_RDWR);
		if (usbfd != -1)
			break;
		if (errno == ENOENT)
			errx(1, "cannot open /dev/usb?");
	}

	/* Find the address of the Compaq iPAQ */
	if (udi.addr == 0) {
		for (udi.addr = 1; udi.addr < USB_MAX_DEVICES; udi.addr++) {
			if (ioctl(usbfd, USB_DEVICEINFO, &udi) == -1)
				continue;
			if (udi.vendorNo == USB_VENDOR_COMPAQ
			    && udi.productNo == USB_PRODUCT_IPAQ)
				break;
		}
		if (udi.addr >= USB_MAX_DEVICES)
			errx(1, "cannot find an iPAQ");
	} else {
		if (ioctl(usbfd, USB_DEVICEINFO, &udi) == -1)
			err(1, "USB_DEVICEINFO addr %u", udi.addr);
	}
	
	/* Find which ugen device is attached to it */
	for (i = 0; i < USB_MAX_DEVNAMES; i++)
		if (memcmp(udi.devnames[i], "ugen", 4) == 0)
			break;
	if (i >= USB_MAX_DEVNAMES)
		errx(1, "cannot find device name");
	close(usbfd);
	ipaqdev = udi.devnames[i];

	/* Open the associated ugen device */
	snprintf(path, sizeof path, "/dev/%s.00", ipaqdev);
	ugenfd = open(path, O_RDWR);
	if (ugenfd == -1)
		err(1, "%s", path);

	/* Show some info about the device */
	if (debug)
		dumpusb(ugenfd);

	/* Open the read & write endpoints */
	snprintf(path, sizeof path, "/dev/%s.01", ipaqdev);
	outfd = open(path, O_WRONLY);
	if (outfd == -1)
		err(1, "%s", path);
	snprintf(path, sizeof path, "/dev/%s.02", ipaqdev);
	infd = open(path, O_RDONLY);
	if (infd == -1)
		err(1, "%s", path);

	/* Allow short transfers on input */
	i = 1;
	if (ioctl(infd, USB_SET_SHORT_XFER, &i) == -1)
		err(1, "USB_SET_SHORT_XFER (infd)");

	/* Run the given program on a new pty, using current stderr */
	errfd = dup(STDERR_FILENO);
	child = forkpty(&pty, NULL, NULL, NULL);
	if (child == -1)
		err(1, "forkpty");
	if (child == 0) {
		close(ugenfd);
		close(infd);
		close(outfd);
		dup2(errfd, STDERR_FILENO);
		close(errfd);
		execvp(argv[optind], argv + optind);
		perror(argv[optind]);
		_exit(1);
	}
	close(errfd);

	/*
	 * Send the two mysterious control messages found
	 * by Ganesh Varadarajan
	 */
	req.data = NULL;
	req.request.bmRequestType = UT_WRITE_CLASS_INTERFACE;
	req.request.bRequest = 0x22;
	USETW(req.request.wValue, 1);
	USETW(req.request.wIndex, 0);
	USETW(req.request.wLength, 0);
	if (ioctl(ugenfd, USB_DO_REQUEST, &req) == -1)
		err(1, "USB_DO_REQUEST (1)");
	if (ioctl(ugenfd, USB_DO_REQUEST, &req) == -1)
		err(1, "USB_DO_REQUEST (2)");

	/* Background */
	daemon(0, 0);

	/* Copy data between the pty and the usb device */
	relay(pty, pty, infd, outfd);

	/* Cleanup */
	close(pty);
	close(infd);
	close(outfd);
	close(ugenfd);
	kill(child, SIGKILL);

	exit(0);
}
